// dijkstra.cpp

#include <iostream>
#include <string>
#include <vector>
#include <fstream>
#include <sstream>
#include <map>
#include <queue>

using namespace std;

class CityEntry {
  public:

      int cityID;
      int distance;
      CityEntry(int,int);
};

CityEntry::CityEntry(int a, int b){
    cityID = a;
    distance = b;
}
bool operator<(const CityEntry & c1, const CityEntry & c2){
  return c1.distance > c2.distance;
}



void printGraph(vector < vector < int > > graph){
  for (auto &row : graph){
    for (auto &column : row){
      cout << column << "\t";
    }
    cout << endl;
    }
}

CityEntry pqTop(vector<CityEntry>  &pq){
  // removes the smallest item in the priority queue
  // and returns it 
  int smallest = pq[0].distance, location = 0;
  for (unsigned int i = 1; i < pq.size(); i++){
    if (pq[i].distance <  smallest){
      smallest = pq[i].distance;
      location = i;
    }
  }
  CityEntry returnCity(pq[location]);
  pq.erase(pq.begin() + location); 
  return returnCity;
}

void pqUpdate(vector<CityEntry> &pq, int id, int d){
  // inputs: id = id of city, d = new distance
  // updates the distance of that city in the priority
  // queue
  for (unsigned int i = 1; i < pq.size(); i++){
    if (pq[i].cityID ==  id){
      pq[i].distance = d;
      break;
    }
  }
}

void Dijkstra(vector < vector < int > > graph, int s, vector<string> cityList){
  vector<CityEntry> pq;
  int length = graph.size();
  vector<int> distance(length, -1);
  vector<int> path(length, -1);
  int v, d;
  CityEntry cE(-1, -1);
  pq.push_back(CityEntry(s, 0));
  distance[s] = 0;
  while (!pq.empty()){
    cE = pqTop(pq);
    v = cE.cityID;
    // for all adjacent vertices 
    for (int w = 0; w < length; w++ ){
      if (graph[v][w] > 0) {
        // compute new distance
        d = distance[v] + graph[v][w];
        if (distance[w] == -1){
          distance[w] = d;
          pq.push_back(CityEntry(w,d));
          path[w] = v;
        }
        else if(distance[w] > d){
          distance[w] = d;
          //update priority queue
          pqUpdate(pq, w, d);
          path[w] = v;
        }

      }
    }

  }

  for (int i = 0; i < length; i++){
      cout << i << ": " << distance[i] << " => " << path[i] << endl;
  }


}


int main(int argc, char **argv) {
  
	string line;
  string sourceCity;
  string destinationCity;
  int distance;
  map<string, int> cities;
  vector<string> cityList;
  int count = 0;
 
  

 	ifstream input(argv[1]);
  // first get cities and determine number of cities
  while (getline(input, line)){

    stringstream lineStream;
    lineStream << line;
    lineStream >> sourceCity;
    cities[sourceCity] = count++;
    cityList.push_back(sourceCity);

  }
  //   for (auto it  = cities.begin(); it != cities.end(); ++it){
  //     cout << it->first << " => " << it->second << endl;
  // }

  // the graph will be represented by a 2d vector
  vector < vector < int > > graph(count, vector<int>(count, -1));
  
  // go back to beginning of file
  input.clear( );              // reset end of file to false
  input.seekg(0, ios::beg);


  // now read in connections between cities
  while(getline(input, line)){
    stringstream lineStream;
    lineStream << line;
    lineStream >> sourceCity;
    int sourceInt = cities[sourceCity];
    graph[sourceInt][sourceInt] = 0;
    while(lineStream >> destinationCity){
      lineStream >> distance;
      if (cities.count(destinationCity) == 0){
        cout << "CANT FIND " << destinationCity << endl;
      }
      graph[sourceInt][cities[destinationCity]] = distance;
    }
  }
    printGraph(graph);
    Dijkstra(graph, 0, cityList);
}

